import math
from bluebeam.objects import Enemy, Laser, HomingRocket
import pygame
from pygame import Vector2 as vec2
import random
import os


# BOSS 2: Beta Spire
# Phase 1: Attacks using homing rockets, bullet barrage, sweeping the screen
# Phase 2: Attacks using homing rockets+, sweeping the screen, laser barrage
# Phase 3 Attacks using sweeping the screen, laser barrage, Air Man knockoff pattern

# There are only two objects making up the boss: this, and the boss body

class BetaSpire(Enemy):
    phase = 3  # phases count in reverse, starts at 4 for the first frame
    phase_tick = 0  # tickdown for phase changes, for the moving animation...
    body = pygame.sprite.Sprite()

    attacks = {
        0: "self.BulletBarrage()",
        1: "self.HomingRockets()",
        2: "self.ScreenSweep()",
        3: "self.LaserBarrage()",
        4: "self.ThatGuyYouCantBeat()",
    }
    attackbursts = (3, 2, 4, 24, 5)

    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._hp = 1
        self._mhp = 1
        self._contactdmg = 0
        self.tpos = pos_
        self.rect = pygame.rect.Rect(self.pos.x, self.pos.y, 72, 72)
        self.rect.center = self.pos
        self.image = None
        self.depth = 32

        self.anchorpos = vec2()
        self._arenabounds = [vec2(), vec2()]
        self._popuppts = [vec2(), vec2(), vec2(), vec2(), vec2()]
        self._nextpopup = 0
        self._popupamt = 960
        self.movemode = 0
        self.waitformove = 0

        self._nextattack = 0 #random.randint(0,1)
        self._attacksuntilretreat = 0
        self._attackcooldown = 1.2
        self._attackburst = 0
        self._turretpoints = [vec2(), vec2(), vec2(), vec2(), vec2(), vec2()]
        self._rocketpoints = [vec2(), vec2()]
        self._jetpoints = [vec2(), vec2(), vec2()]
        self._sweep = 0

        self._sprites = []
        self._glow = 0.0
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imagedir = os.path.join(os.path.dirname(self.sourceDir), "images/Boss/BetaSpire")
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss02_core.png')).convert()
        self.load_sprites()
        self.playerdir = 1

        self.phase_change()
        pass

    def load_sprites(self):
        for image in range(4):
            x = 24 * (image % 2)
            y = 24 * math.floor(image / 2)
            img = self.load_single((x, y, 24, 24))
            self._sprites.append(pygame.transform.scale(img, (72, 72)))
        self.image = self._sprites[0]

    @property
    def hp(self):
        return self._hp

    @hp.setter
    def hp(self, hp_):
        self._hp = hp_
        if self._hp <= 0 and self.phase > 0:
            self.phase_down()

    def setup(self, arena_lt, arena_rb):
        self.body = self.l.create("SpireBody", self.pos.xy)
        self.body._bosscore = self
        self._anchorpos = (arena_lt + arena_rb) / 2
        self._arenabounds = [arena_lt, arena_rb]
        self._nextpopup = 1
        for i in range(0, 6):
            self._turretpoints[5 - i] = vec2(self.body.rect.width / 2,
                                             128 + (i * (self.body.rect.height - 160) / 5 - self.body.rect.height / 2))
        for i in range(0, 2):
            self._rocketpoints[i] = vec2(self.body.rect.width / 2 + -72 + 72 * i, -156)
        for i in range(0, 3):
            self._jetpoints[i] = (self._turretpoints[2*i] + self._turretpoints[2*i+1])/2
        self.body.movement()
        self.determinepopups()
        self._attackcooldown = 1.2
        self._nextattack = 1
        self.body.activeextents = self._nextattack
        self._attackburst = self.attackbursts[self._nextattack]

    def determinepopups(self):
        h = 2
        space = ((self._arenabounds[1].x - self._arenabounds[0].x) - self.body.rect.width) / (h - 1)
        for i in range(0, h):
            self._popuppts[i] = vec2(self._arenabounds[0].x + self.body.rect.width / 2 + space * i,
                                     self._arenabounds[1].y - 284)

    def movement(self):
        if self.movemode == 0:
            self._popupamt = max(0, self._popupamt - 720 * self.g.delta_time)
        elif self.movemode == 1:
            self._popupamt = min(960, self._popupamt + 720 * self.g.delta_time)
        self.pos = self._popuppts[self._nextpopup] + vec2((self._sweep * self.playerdir), self._popupamt)
        self.rect.center = self.pos
        if self._sweep == 0:
            self.playerdir = math.copysign(1, self.l.player.pos.x - self.pos.x)
        self.body.movement()

    def update(self):
        if self._active:
            if self.phase > 0:
                self.spire_logic()
                self.collision_bullet()
            else:
                self.fadetimer = max(0, self.fadetimer - self.g.delta_time)
                if self.fadetimer < 4.5:
                    self.rect.center = self.pos = self.pos + vec2(0,300) * self.g.delta_time
                    self.body.movement()
                if self.fadetimer == 0:
                    self.death()
        else:
            self.check_active()

    def spire_logic(self):
        if self.waitformove == 0 or self._nextattack == 2:
            self.movement()

        if self.movemode == 0 and self._popupamt == 0 and self.waitformove == 0:
            self.waitformove = 1.2
            self._attackcooldown = 1.2
            self._attackburst = self.attackbursts[self._nextattack]

        elif self.movemode == 1 and self._popupamt == 960 and self.waitformove == 0:
            self.waitformove = 1.2
            self._attackcooldown = 0
            self._nextpopup = random.randint(0, 1)
            self._nextattack = random.randint(0, 2) + (3-self.phase)
            self.body.activeextents = self._nextattack
            self._sweep = 0

        if self._attackcooldown == 0 and self._attackburst == 0 and self.waitformove > 0:
            self.waitformove = max(0, self.waitformove - self.g.delta_time)
            self.body.activeextents = self._nextattack
            self.body.extentoffset = (self.movemode == 0) * max(0.0, self.waitformove - 0.4) / 0.8
            if self.waitformove == 0:
                self.movemode = 1 - self.movemode
                if (self.movemode == 0):
                    self._attackburst = self.attackbursts[self._nextattack]

        elif self.movemode == 0 and self._popupamt == 0:
            self._attackcooldown = max(0, self._attackcooldown - self.g.delta_time)
            if self._attackburst == self.attackbursts[self._nextattack]:
                self.body.extentoffset = 1 - (max(0.0, self._attackcooldown - 0.4) / 0.8)
            if self._attackcooldown == 0 and self._attackburst > 0:
                eval(self.attacks[self._nextattack])

    def death(self):
        if self.deathstate == 1 and self.fadetimer == 0:
            self.l.bossDeathState == 1
            self.l.uncreate(self.instid)

    def phase_down(self):
        self.phase -= 1
        # fallback item
        for eb in self.l.bulletsE:
            if isinstance(eb, Laser):
                eb.fadeout()
            else:
                eb.hit()
        if self.phase <= 0:
            self.body.phasedownexplo = 80
            self.deathstate = 1
            self.fadetimer = 5.5
            pygame.mixer.music.fadeout(1800)
        else:
            self.body.phasedownexplo = 12
            self.waitformove = 0.8
            self._attackburst = 0
            self.phase_change()

    def phase_change(self):
        if self.phase == 3:
            self._mhp = 200
        elif self.phase == 2:
            self._mhp = 250
            pass
        elif self.phase == 1:
            self._mhp = 300
        self._hp = self._mhp

    def attack(self):
        pass

    def BulletBarrage(self):
        if self._attackburst == self.attackbursts[0]:
            las = self.l.create("Laser", vec2(self.body.pos.x + 24 + 2 + 48 * self.playerdir, self.body.pos.y - 160))
            las.width = 48
            las.dir = 1 - self.playerdir
            las.lifespan = 1.75
        patterns = [
            (0, 5),
            (0, 3),
            (0, 1),
            (0, 1, 2),
            (0, 1, 5),
            (0, 1, 4, 5),
            (1, 2),
        ]
        select = random.randint(0, len(patterns) - 1)
        for i in patterns[select]:
            bul = self.l.create("EnemyBullet", self.body.pos + self._turretpoints[i])
            bul.vel = vec2(self.playerdir * 800, 0)
        self._attackburst -= 1
        self._attackcooldown = 0.75
        pass

    def HomingRockets(self):
        for i in range(0, 2):
            hom = self.l.create("HomingRocket", self.body.pos + self._rocketpoints[i])
        self._attackburst -= 1
        self._attackcooldown = 1.6
        pass

    def ScreenSweep(self):
        if (self._attackburst > 1):
            self._attackburst -= 1
            self._attackcooldown = 1 - 0.2*(3-self.phase)
            pass
        else:
            self._sweep += self.g.delta_time * 1350
            self.body.pushforce = 0
            self.body.contactdmg = 20
            if self._sweep > self._arenabounds[1].x - self._arenabounds[0].x:
                self.body.pushforce = 960
                self.body.contactdmg = 10
                self._attackburst = 0
                self._attackcooldown = 0
                self.waitformove = 0.1
            pass
        pass

    def LaserBarrage(self):
        if self._attackburst > 17:
            las = self.l.create("Laser",
                                vec2(self.body.pos.x + (128 + 160 * (24 - self._attackburst)) * self.playerdir,
                                     self.body.pos.y - 512))
            las.width = 48
            las.dir = 1
            las.warmup = 0.67
            las.lifespan = 0.3
            self._attackcooldown = 0.14
        elif 17 >= self._attackburst > 10:
            las = self.l.create("Laser",
                                vec2(self.body.pos.x + (64 + 160*7 - 160 * (17 - self._attackburst)) * self.playerdir,
                                     self.body.pos.y - 512))
            las.width = 48
            las.dir = 1
            las.warmup = 0.67
            las.lifespan = 0.3
            self._attackcooldown = 0.12
        elif 10 >= self._attackburst > 3:
            las = self.l.create("Laser",
                                vec2(self.body.pos.x + (128 + 160 * (10 - self._attackburst)) * self.playerdir,
                                     self.body.pos.y - 512))
            las.width = 48
            las.dir = 1
            las.warmup = 0.6
            las.lifespan = 0.3
            self._attackcooldown = 0.09
        elif self._attackburst > 0:
            alt = (abs(self._attackburst - 1) % 2)
            las = self.l.create("Laser",
                                vec2(self.body.pos.x + (128 + 992 * alt) * self.playerdir,
                                     self.body.pos.y - 512))
            las.width = 96
            las.dir = 1
            las.warmup = 1.0
            las.lifespan = 1.25
            las.b2speed = (450 * self.playerdir * (1 - 2*alt))
            self._attackcooldown = 1.8
        self._attackburst -= 1

    def ThatGuyYouCantBeat(self):
        if self._attackburst == self.attackbursts[0]:
            las = self.l.create("Laser", vec2(self.body.pos.x + 24 + 2 + 48 * self.playerdir, self.body.pos.y - 160))
            las.width = 48
            las.dir = 1 - self.playerdir
            las.lifespan = 2.4
            las.warmup = 0.8
        patterns = [
            (0, 10, 12),
            (5, 6, 9, 13),
            (15, 7, 0),
        ]
        select = random.randint(0, len(patterns) - 1)
        for i in patterns[select]:
            bul = self.l.create("SuperEnemyBullet", self.body.pos + vec2(self._turretpoints[(i%6)].x * self.playerdir, self._turretpoints[(i%6)].y))
            bul.targets = [self.body.pos + vec2((192 + 128*math.floor(i/6))*self.playerdir, 0)+ vec2(0,self._turretpoints[(i%6)].y), self.body.pos + vec2(3072*self.playerdir, 0)+ self._turretpoints[i%6]]
        self._attackburst -= 1
        self._attackcooldown = 0.8
        pass

    def death(self):
        self.l.uncreate(self.body.instid)
        self.l.uncreate(self.instid)
        self.l.bossDeathState = 1

    def render(self):
        if self._active:
            self._glow += 7.5 * self.g.delta_time
            if (self._glow >= 7):
                self._glow -= 7
            self.image = self._sprites[3 - int(abs(3 - self._glow))]
            self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))
            if self.phase > 0 and self._active:
                # Draw boss health bar at bottom of screen
                sz = self.g.screen_size
                mid = sz[0] / 2 - 490
                top = 80
                hel = float(self._hp / self._mhp) * 968
                pygame.draw.rect(self.l.s, (128, 0, 96, 0), pygame.rect.Rect(mid, top, 980, 60))
                pygame.draw.rect(self.l.s, (255, 0, 192, 100), pygame.rect.Rect(mid + 6, top + 6, hel, 48))



import math

from bluebeam.objects import PhysObject, PlayerBullet, YellowPower, BossSpawnTrigger
import pygame
import os


class Player(PhysObject):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._hp = 100
        self.recently_damaged = False
        self.speed = pygame.math.Vector2(500, 1000)
        # self.rect = pygame.rect.Rect(0, 0, 100, 100) ORIGINAL
        self.rect = pygame.rect.Rect(0, 0, 24, 64)
        self.rect.center = self.pos
        self.shotcd = 0
        self.confuse_cd = 0
        self.invuln = 1.2
        self.gravity = 50
        self.jumping = False
        self.charge_unlocked = False
        self.jump_gravity_mod = 0.3  # Gravity is only x% effective while holding space!

        # for left right movement
        self.running_left = []

        self.stand_pos = 0
        # Load sprite sheet
        self.idle = []
        self.jump = []
        self.falling = []
        self.running_right = []
        self.arm = []
        self.activearm = None
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images")
        self.soundDir = os.path.join(os.path.dirname(self.sourceDir), "Sounds")
        self.spritesheet = pygame.image.load(os.path.join(self.imageDir, 'sigma.png')).convert()
        self.chroma = -1
        self.load_sprites()

        self.image = self.running_right[self.stand_pos]
        self.confusion_img = pygame.transform.scale(pygame.image.load(os.path.join(self.imageDir,"Confusion.png")).convert_alpha(), (44,32))
        #self.rect = self.image.get_rect()
        #self.width = self.image.get_width()
        #self.height = self.image.get_height()
        self.animation_counter = 0
        self.direction = 1
        self.aimdirection = pygame.math.Vector2(0, 0)

        # power-ups changes bullet. Default will be zero
        # will decide soon if the power-ups will come from items found or
        # how far in the level player has traversed/enemies killed/score?
        self.current_power = ""
        self.power_up = 0
        self.power_changed = False
        self.charged = False
        self.charged_counter = 0

        #loading some sounds
        self.powup_sound = pygame.mixer.Sound(os.path.join(self.soundDir, 'up.wav'))
        self.powup_sound.set_volume(0.5 * (self.g.get_music_vol() / 100))
        self.charging_sound = pygame.mixer.Sound(os.path.join(self.soundDir, 'charge.wav'))
        self.charging_sound.set_volume(0.5 * (self.g.get_music_vol() / 100))
        self.shoot_sound = pygame.mixer.Sound(os.path.join(self.soundDir, 'shot.wav'))
        self.shoot_sound.set_volume(0.5 * (self.g.get_music_vol() / 100))
        self.charged_sound = pygame.mixer.Sound(os.path.join(self.soundDir, "charged.wav"))
        self.charged_sound.set_volume(0.5 * (self.g.get_music_vol() / 100))
        self.cshot_sound = self.shoot_sound
        self.cshot_sound.set_volume(0.5 * (self.g.get_music_vol() / 100))
        self.damaged_sound = pygame.mixer.Sound(os.path.join(self.soundDir, 'damaged.wav'))
        self.damaged_sound.set_volume(0.5 * (self.g.get_music_vol() / 100))

        #for littlegrey default attack is telepath (inverts player controls)
        self.around_lg = False

        self.power_yellow = 0
        self.power_blue = 0
        self.continuous_shot = False
        self.shooting = False
    def is_player_out_of_bounds(self):
        if self.pos.y > 2528:
            self._hp = 0

    def turn_player_red(self):
        self.idle.clear()
        self.jump.clear()
        self.falling.clear()
        self.running_right.clear()
        self.running_left.clear()
        self.arm.clear()
        self.spritesheet = pygame.image.load(os.path.join(self.imageDir, 'sigmaRed.png')).convert()
        self.load_sprites("red")

    def turn_player_green(self):
        self.idle.clear()
        self.jump.clear()
        self.falling.clear()
        self.running_right.clear()
        self.running_left.clear()
        self.arm.clear()
        current_arm = self.activearm
        self.spritesheet = pygame.image.load(os.path.join(self.imageDir, 'sigma.png')).convert()
        self.load_sprites("green")
        self.activearm = current_arm

    @property
    def hp(self):
        return self._hp
    @hp.setter
    def hp(self, amount):
        self._hp = amount

    def load_sprites(self, arm_color="green"):
        # Standing sprite
        self.idle.append(self.load_single((0,0,64,64)))

        # Jumping sprite
        self.jump.append(self.load_single((64, 0, 64, 64)))

        # Running sprites
        for image in range(10):
            x = 64 * (image % 5)
            y = 64 + 64 * math.floor(image/5)
            img = self.load_single((x,y,64,64))
            imgflip = pygame.transform.flip(img, True, False)
            self.running_right.append(img)
            self.running_left.append(imgflip)

        # Load arms
        if arm_color == "green":
            tempsheet = pygame.image.load(os.path.join(self.imageDir, "arm.png")).convert()
        elif arm_color == "red":
            tempsheet = pygame.image.load(os.path.join(self.imageDir, "armRed.png")).convert()
        #tempsheet = pygame.image.load(f'images/arm.png').convert()
        rect = pygame.Rect(80, 16, 40, 11)
        img = pygame.Surface(rect.size).convert()
        img.blit(tempsheet, (0, 0), rect)
        img.set_colorkey(img.get_at((0, 0)), pygame.RLEACCEL)
        self.arm.append(img)
        rect = pygame.Rect(80, 16, 40, 11)
        img = pygame.Surface(rect.size).convert()
        img.blit(tempsheet, (0, 0), rect)
        img.set_colorkey(img.get_at((0, 0)), pygame.RLEACCEL)
        self.arm.append(img)
        self.activearm = self.arm[1]

    # Returns how much the player's shot is charged up, as a percentage
    def get_charge_level(self):
        if(self.charged): return 1.0
        else:
            return float(self.charged_counter) / 50.
    
    def get_powerup_level(self):
        return self.power_up
            
    def update(self):
        self.check_inputs()
        self.update_sprite()
        self.power_sprites()

        gravity_mod = 1
        if (self.jumping):
            gravity_mod = self.jump_gravity_mod

        self.update_phys(gravity_mod)
        self.collision_bullet()
        self.collision_trigger()
        self.shotcd = max(self.shotcd - self.g.delta_time, 0)
        if(self.invuln > 0):
            self.invuln = max(self.invuln - self.g.delta_time, 0)
            if(self.invuln <= 0):
                self.turn_player_green()

        self.shoot_sound.set_volume(0.5 * (self.g.get_FX_vol() / 100))

    # Read and manage all the inputs for player
    def check_inputs(self):
        keys = pygame.key.get_pressed()
        #if not self.l.check_grey():
        if not self.around_lg:
            self._vel.x = (keys[pygame.K_d] - keys[pygame.K_a]) * self._speed.x
        else:
            self._vel.x = -(keys[pygame.K_d] - keys[pygame.K_a]) * self._speed.x

        # Jumping
        if (keys[pygame.K_SPACE]):
            if (self.pstate == 0):
                self.jumping = True
                self._vel.y = -self.speed.y
        elif (self.jumping == True):
            self.jumping = False



        # Charging
        if keys[pygame.K_RSHIFT] and not self.charged and self.charge_unlocked:
            self.charged_counter += 1
            #place charging sound here
            #self.charging_sound.play()

            # print(self.charged_counter)
        if self.charged_counter >= 50:
            self.charged_sound.play()
            self.charged = True
            self.charged_counter = 0


        # Shooting
        if not self.around_lg:
            self.aimdirection = pygame.math.Vector2(keys[pygame.K_RIGHT] - keys[pygame.K_LEFT], keys[pygame.K_DOWN] - keys[pygame.K_UP])
        else:
            self.aimdirection = -1 * pygame.math.Vector2(keys[pygame.K_RIGHT] - keys[pygame.K_LEFT], keys[pygame.K_DOWN] - keys[pygame.K_UP])

        if self.aimdirection.length() > 0:
            if self.power_yellow > 0:
                self.shoot(self.aimdirection.x, self.aimdirection.y)

            elif self.power_yellow == 0 and not self.shooting:
                self.shoot(self.aimdirection.x, self.aimdirection.y)
                self.shooting = True
        else:
            self.shooting = False

        # elif self.aimdirection.length() > 0 and self.power_yellow == 0 and self.shooting:
        #     self.shooting = False
        #     #print("does this print")

    def power_sprites(self):
    # obtaining powerUp(s)
        keys = pygame.key.get_pressed()
        power = pygame.sprite.spritecollide(self, self.l.powerUps, False)
        if len(power) > 0:
            for booster in power:
                #old_power = self.power_up
                boost_type = booster.acquired()
                self.powup_sound.play()
                booster.kill()

                if  (boost_type == 1):
                    self.current_power = "blue"
                    self.charge_unlocked = True
                    self.power_blue += 1
                elif(boost_type == 2):
                    self.current_power = "yellow"
                    self.power_up += 1
                    self.power_yellow += 1
                elif(boost_type == 3):
                    self.current_power = "red"
                    self.hp += 10
                    if self.hp > 100:
                        self.hp = 100

    # Update the Player's Animation
    def update_sprite(self):
        self.is_player_out_of_bounds()

        if self._vel.x != 0: self.direction = math.copysign(1, self._vel.x)

        self.truedir = self.direction
        if (self.aimdirection.x != self.direction) and self.aimdirection.x != 0:
            self.truedir = self.aimdirection.x

        if self._vel.x < 0:
            self.animation_counter += 18 * self.g.delta_time

            if self.animation_counter >= 10:
                self.animation_counter = 0
        elif self._vel.x > 0:
            self.animation_counter += 18 * self.g.delta_time

            if self.animation_counter >= 10:
                self.animation_counter = 0
        else:
            self.animation_counter = 0

        if abs(self._vel.y) > self.gravity:
            self.image = self.jump[0]
        elif self._vel.x != 0:
            self.image = self.running_right[math.floor(self.animation_counter)]
        else:
            self.image = self.idle[0]

        self.activearm = self.arm[1]

        if (self.truedir == -1):
            self.image = pygame.transform.flip(self.image, True, False)
            self.activearm = pygame.transform.flip(self.activearm, True, False)

    def shoot(self, lr, ud):
        if self.shotcd <= 0:
            inst = self.l.create("PlayerBullet", self.pos.xy)
            inst.speed = pygame.math.Vector2(lr * 800, ud * 800)
            if (self.aimdirection.x == 1 and self.aimdirection.y == 0 and not self.charged):
                inst.pos.x += 18
                inst.pos.y -= 10
            elif (self.aimdirection.x == -1 and self.aimdirection.y == 0 and not self.charged):
                inst.pos.x -= 18
                inst.pos.y -= 10
            elif (self.aimdirection.x == 1 and self.aimdirection.y == 0 and self.charged):
                inst.pos.x += 30
                inst.pos.y -= 10
            elif (self.aimdirection.x == -1 and self.aimdirection.y == 0 and self.charged):
                inst.pos.x -= 30
                inst.pos.y -= 10
            elif (self.aimdirection.x == 0 and self.aimdirection.y == -1 and not self.charged):
                inst.pos.y -= 30
            elif (self.aimdirection.x == 0 and self.aimdirection.y == -1 and self.charged):
                inst.pos.y -= 40
            elif (self.aimdirection.x == 1 and self.aimdirection.y == -1 and not self.charged):
                inst.pos.x += 15
                inst.pos.y -= 22
            elif (self.aimdirection.x == 1 and self.aimdirection.y == -1 and self.charged):
                inst.pos.x += 30
                inst.pos.y -= 22

            elif (self.aimdirection.x == -1 and self.aimdirection.y == -1 and not self.charged):
                inst.pos.x -= 12
                inst.pos.y -= 22
            elif (self.aimdirection.x == -1 and self.aimdirection.y == -1 and self.charged):
                inst.pos.x -= 30
                inst.pos.y -= 35

            if self.charged:
                self.cshot_sound.play()
                inst.dmg *= 1.5 * self.power_blue
                inst.size = 64
                self.charged = False
                inst.charged = True

            if self.power_yellow >= 1:
                self.shotcd = .4 - self.power_yellow * .05
                self.shoot_sound.play()
            else:
                self.shotcd = .4
                self.shoot_sound.play()

    def rotinplace(self, image, angle):
        core = image.get_rect().center
        ret = pygame.transform.rotate(image, angle)
        rec = ret.get_rect(center = ret.get_rect(center = (core[0],core[1])).center)
        return ret, rec

    def render(self):
        # Draw arm
        if self.aimdirection == 0 and self._vel.length() == 0:
            self.l.s.blit(self.activearm, self.rect.copy().move(-self.l.view).move(pygame.math.Vector2(12 + 3*self.direction,15)))
        else:
            inv = (self.truedir == -1 and self.aimdirection.x != 0)
            ang = 180*(self.truedir==-1 and self.aimdirection.x != 0) + pygame.math.Vector2(self.aimdirection.x,-self.aimdirection.y).as_polar()[1]
            rarm = self.rotinplace(self.activearm, ang)
            self.l.s.blit(rarm[0], rarm[1].copy().move(-self.l.view + self.pos - pygame.math.Vector2(30)).move(pygame.math.Vector2(10,15)))
        # Draw player
        #pygame.draw.rect(self.l.s, (255,255,0,100),self.rect.copy().move(-self.l.view))
        self.l.s.blit(self.image, self.rect.copy().move(-self.l.view - pygame.math.Vector2(20,0)))
        if(self.around_lg):
            self.l.s.blit(self.confusion_img, self.rect.copy().move(-self.l.view - pygame.math.Vector2(10,50)))


    def collision_bullet(self):
        col = pygame.sprite.spritecollide(self, self.l.bulletsE, False)
        if len(col) > 0:
            for bul in col:
                self.take_damage(bul.dmg)
                bul.hit()

    def collision_trigger(self):
        col = pygame.sprite.spritecollide(self, self.l.triggers, False)
        if len(col) > 0:
            for bul in col:
                if bul.activated != 1:
                    bul.firstframe()


    def take_damage(self, damage):
        if self.invuln <= 0 and damage > 0:
            self._hp -= damage
            self.invuln = 0.5
            self.damaged_sound.play()
            self.turn_player_red()



from bluebeam.objects import WhaleTurret
import pygame
import math


class WhaleTurretSide(WhaleTurret):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._turretpoint = self.pos
        self._extenddelay = 4.5
        self._extend = 0

    def update(self):
        self.style = 1
        super().update()
        self._turretpoint = self.pos
        self.pos = self._turretpoint + pygame.math.Vector2(0, self._extend)
        self._extenddelay -= self.g.delta_time
        if self._shotcd > 0.8 and self._extenddelay <= 0:
            value = self.l.player.pos.y - (self._turretpoint.y + self._extend)
            self._extend += math.copysign(min(abs(value), 175 * self.g.delta_time), value)

    def fire(self):
        inst = self.l.create("EnemyBullet", self._turretpoint + pygame.math.Vector2(0, self._extend))
        inst.speed = pygame.math.Vector2(400 * math.cos(math.radians(self._rot)),
                                         400 * math.sin(math.radians(self._rot)))
        self._shotcd = 4.33

    def render(self):
        super().render()



# Generics
from .GameObject import GameObject
from .PhysObject import PhysObject
from .Bullet import Bullet
from .Block import Block
from .Background import Background
# Player
from .Player import Player
from .PlayerBullet import PlayerBullet
# Enemies
from .Enemy import Enemy
from .Target import Target
from .BadTarget import BadTarget
from .SliderEnemy import SliderEnemy
from .ShootyEnemy import ShootyEnemy
from .BulkyEnemy import BulkyEnemy
# Enemy Projectiles
from .EnemyBullet import EnemyBullet
from .SuperEnemyBullet import SuperEnemyBullet
from .Rocket import Rocket
from .HomingRocket import HomingRocket
from .Laser import Laser
from .SplitBomb import SplitBomb
from .WaveBomb import WaveBomb
from .Shockwave import Shockwave
from .Explosion import Explosion
from .Tentacle import Tentacle
# Boss 1 Parts
from .WhaleTurret import WhaleTurret
from .WhaleCore import WhaleCore
from .WhaleShield import WhaleShield

from .WhaleCommander import WhaleCommander
from .WhaleTurretFlex import WhaleTurretFlex
from .WhaleTurretSide import WhaleTurretSide
from .WhaleTurretRocket import WhaleTurretRocket
from .WhaleTurretLaser import WhaleTurretLaser

from .AlphaWhale import AlphaWhale

# Boss 2
from .SpireBody import SpireBody
from .BetaSpire import BetaSpire
# Boss 3
from .GammaGenerator import GammaGenerator
from .GammaHand import GammaHand
from .GammaHead import GammaHead
from .GammaBGRocket import GammaBGRocket
from .GammaCore import GammaCore
# Triggers
from .BossSpawnTrigger import BossSpawnTrigger
from .BetaBossSpawnTrigger import BetaBossSpawnTrigger
from .GammaBossSpawnTrigger import GammaBossSpawnTrigger

from .LittleGrey import LittleGrey

#PowerUps
from .YellowPower import YellowPower
from .BluePower import BluePower
from .RedPower import RedPower



from bluebeam.objects import Enemy
import pygame, math, random, os
from pygame import Vector2 as vec2

class GammaHead(Enemy):
    def __init__(self, g, l, pos):
        super().__init__(g, l, pos)
        self.rect = pygame.rect.Rect(0, 0, 832, 288)
        self.body = None
        self.handtype = vec2(1, 1)
        self.sprite = None

        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imagedir = os.path.join(os.path.dirname(self.sourceDir), "images/Boss/GammaGenerator")
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss03_head.png')).convert()
        self.load_sprites()

        self.neckpoint = vec2()

        self.depth = 10

        self.mode = 0
        self.stateclock = 0
        self.bob = 0

        self.moveto = vec2()
        self.waveburst = 0

    def load_sprites(self):
        img = self.load_single((0,0,208,72))
        self.sprite = pygame.transform.scale(img, (832,288))

    def setup(self, body_):
        self.body = body_
        self.neckpoint = self.pos

    def state(self, mode_):
        self.mode = mode_
        self.stateclock = 0
        if mode_ == 1:
            self.waveburst = 0
            self.wave_reposition()

    def update(self):
        self.process()

    def process(self):
        pass
        states = {
            0: self.idle,
            1: self.wave
        }
        states[self.mode]()

    def smooth_move(self, target_, speed):
        mag = (target_ - self.pos).magnitude()
        if mag > 0:
            if (target_ - self.pos).magnitude() < math.pow(mag, 0.45):
                self.rect.center = self.pos = target_
            else:
                amt = min(speed*self.g.delta_time, math.pow(mag, 0.45))
                self.rect.center = self.pos = self.pos + amt * (target_ - self.pos).normalize()

    def idle(self):
        self.bob = (self.bob + self.g.delta_time * 180) % 360
        self.smooth_move(self.body.pos + vec2(0, -128), 300)

    def wave(self):
        if self.pos != self.moveto:
            self.smooth_move(self.moveto, 600)
            self.stateclock = 0
        else:
            self.stateclock += self.g.delta_time
            if self.stateclock > 0.5 - 0.1*(3-self.body.phase):
                self.l.create("WaveBomb", self.pos.xy)
                self.waveburst += 1
                if self.waveburst >= 3 + (3-self.body.phase):
                    self.state(0)
                else:
                    self.wave_reposition()

    def wave_reposition(self):
        self.rect.center = self.pos
        self.moveto = vec2(random.randint(self.body.bounds[0].x + 144, self.body.bounds[1].x - 144), self.body.pos.y - 160)

    def wave_spit(self):
        self.l.create("WaveBomb", self.pos.xy)

    def death(self):
        pass

    def render(self):
        self.l.s.blit(self.sprite,self.rect.copy().move(-self.l.view))


from bluebeam.objects import EnemyBullet, Explosion
import pygame


class Rocket(EnemyBullet):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.lifespan = 10.
        self.sprite = pygame.sprite.Sprite()

    def rotinplace(self, image, angle):
        core = image.get_rect().center
        ret = pygame.transform.rotate(image, angle)
        rec = ret.get_rect(center=ret.get_rect(center=(core[0], core[1])).center)
        return ret, rec

    def hit(self):
        boom = self.l.create("Explosion", self.pos.xy)
        boom.radius(192)
        self.l.uncreate(self.instid)



import itertools

import pygame
import math
from bluebeam.objects import EnemyBullet


class Laser(EnemyBullet):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.owner = 1
        self._emitter = None
        self.emitpoint = self.pos
        self.speed = pygame.math.Vector2(0, 0)
        self.vel = self.speed
        self.warmup = 0.5
        self.lifespan = 2.33
        self._objid = 4
        self._dir = 1
        self.b2speed = 0
        self._width = 16
        self.dmg = 0
        self.times = -1

    @property
    def dir(self):
        return self._dir

    @dir.setter
    def dir(self, d_):
        self._dir = d_

    @property
    def emitter(self):
        return self._emitter

    @emitter.setter
    def emitter(self, e_):
        self._emitter = e_

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, w_):
        self._width = w_

    def expand(self):
        self.times -= 1
        len = 16
        pow = 6
        wid = (self.warmup <= 0)*(self.width-2) + 2
        self.rect = pygame.rect.Rect(self.emitpoint.x, self.emitpoint.y, 2 + 14 * (self._dir % 2 == 0),
                                     2 + 14 * (self._dir % 2 == 1))
        if abs(self.b2speed) != 0 and self.warmup <= 0:
            self.emitpoint.x += self.b2speed * self.g.delta_time
        self.set_dir()
        while pow >= 0 and len < 1024:
            len += math.pow(4, pow)
            self.rect = self.rect.inflate(math.pow(4, pow) * (self._dir % 2 == 0),
                                          math.pow(4, pow) * (self._dir % 2 == 1))
            self.set_dir()
            if self.collision_surface():
                self.rect = self.rect.inflate(-math.pow(4, pow) * (self._dir % 2 == 0),
                                              -math.pow(4, pow) * (self._dir % 2 == 1))
                self.set_dir()
                len -= math.pow(4, pow)
                pow -= 1
        self.rect = self.rect.inflate((wid - 2) * (self._dir % 2 == 1),
                                      (wid - 2) * (self._dir % 2 == 0))

    def set_dir(self):
        if self._dir == 0:
            self.rect.midleft = self.emitpoint
        elif self._dir == 1:
            self.rect.midtop = self.emitpoint
        elif self._dir == 2:
            self.rect.midright = self.emitpoint
        elif self._dir == 3:
            self.rect.midbottom = self.emitpoint

    def update(self):
        if self.times >= 0:
            self.killproj()
        if  self._emitter is not None:
            self.emitpoint = self._emitter.emitpoint
        if self.times != 0:
            self.expand()
        if self.warmup > 0:
            self.warmup = max(self.warmup - self.g.delta_time, 0)
            if self.warmup == 0:
                self.dmg = 10
        else:
            self.lifespan -= self.g.delta_time
            if self.lifespan <= 0: self.fadeout()

    def fadeout(self):
        self.l.uncreate(self.instid)

    def hit(self):
        pass

    def killproj(self):
        col = itertools.chain(pygame.sprite.spritecollide(self, self.l.bulletsE, False), pygame.sprite.spritecollide(self, self.l.bulletsP, False))
        for bul in col:
            bul.hit()

    def render(self):
        pygame.draw.rect(self.l.s, (64 + 191*(self.times >= 0), 64, 255, 100), self.rect.copy().move(-self.l.view))



import pygame
import math, os
from bluebeam.objects import EnemyBullet


class Tentacle(EnemyBullet.EnemyBullet):
    def __init__(self, g_, l_, tentacle_num, left, pos_=None):
        self.pull_tentacle_info(tentacle_num, left)
        true_pos = pos_ + self.offset
        super().__init__(g_, l_, true_pos)

        self._dmg = 20
        self.speed = pygame.math.Vector2(0., 0.)
        self.vel = self.speed
        self.lifespan = 300
        self._from_bulky = True


    def pull_tentacle_info(self, tentacle_num, left):
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imageDir = os.path.join(os.path.dirname(self.sourceDir), "images/Enemy/BulkyGuy")
        if(tentacle_num == 4):
            self.image = pygame.transform.scale( pygame.image.load(os.path.join(self.imageDir, "ArmsAttack4.png")).convert_alpha() , (194, 166))
            self.rect = pygame.rect.Rect(0,0,194, 166)
            self.offset = pygame.math.Vector2(100,-100)
        elif(tentacle_num == 5):
            self.image = pygame.transform.scale( pygame.image.load(os.path.join(self.imageDir, "ArmsAttack5.png")).convert_alpha() , (160, 112))
            self.rect = pygame.rect.Rect(0,0,160, 112)
            self.offset = pygame.math.Vector2(150,40)
        elif(tentacle_num == 6):
            self.image = pygame.transform.scale( pygame.image.load(os.path.join(self.imageDir, "ArmsAttack6.png")).convert_alpha() , (142, 46))
            self.rect = pygame.rect.Rect(0,0,142, 46)
            self.offset = pygame.math.Vector2(140,100)
            if(not left): self.offset.x += 20
        elif(tentacle_num == 7):
            self.image = pygame.transform.scale( pygame.image.load(os.path.join(self.imageDir, "ArmsAttack7.png")).convert_alpha() , (94, 76))
            self.rect = pygame.rect.Rect(0,0,94, 76)
            self.offset = pygame.math.Vector2(85,35)
            if(not left): self.offset.x += 65
        else:
            self.image = None
            self.offset = pygame.math.Vector2(0,0)

        if(left): 
            self.image = pygame.transform.flip(self.image, True, False)
            self.offset.x = -(self.offset.x + 150)
        else:
            self.offset.x += 55

    def refresh_pos(self, new_pos):
        true_pos = new_pos + self.offset
        self._pos = true_pos

    def render(self):
        spos = self._pos - self.l.view
        self.l.s.blit(self.image, self.rect.copy().move(-self.l.view))

    


from bluebeam.objects import WhaleTurret
import pygame
import math


class WhaleTurretFlex(WhaleTurret):
    def __init__(self, g_, l_, pos_=None):
        self.style = 0
        super().__init__(g_, l_, pos_)

    def update(self):
        super().update()
        self._rot = pygame.math.Vector2().angle_to(self.l.player.pos - self.pos)

    def fire(self):
        inst = self.l.create("EnemyBullet", self.pos.xy)
        inst.speed = pygame.math.Vector2(300 * math.cos(math.radians(self._rot)),
                                         300 * math.sin(math.radians(self._rot)))
        self._shotcd = 2.67

    def render(self):
        super().render()



from bluebeam.objects import PhysObject
import pygame


class Enemy(PhysObject):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self._objid = 2
        self._hp = 1
        self._contactdmg = 10
        self._active = False
        self._telepath = False
        pass

    def update(self):
        if(self._active):
            self.update_phys()
            self.collision_bullet()
            self.collision_player()
        else:
            self.check_active()


    @property
    def hp(self):
        return self._hp

    @hp.setter
    def hp(self, hp_):
        self._hp = hp_
        if self._hp <= 0:
            self.death()

    @property
    def contactdmg(self):
        return self._contactdmg

    @contactdmg.setter
    def contactdmg(self, contactdmg_):
        self._contactdmg = contactdmg_

    def death(self):
        self.l.uncreate(self.instid)

    def check_active(self):
        diff = self.pos - self.l.player.pos
        squareddist = diff.x**2 + diff.y**2
        if squareddist < 1100000:
            self._active = True

    def collision_bullet(self):
        col = pygame.sprite.spritecollide(self, self.l.bulletsP, False)
        if len(col) > 0:
            for bul in col:
                self.hp -= bul.dmg
                if(bul.charged):
                    self.vel += bul.vel
                bul.hit()

    def collision_player(self):
        col = pygame.sprite.collide_rect(self, self.l.player)
        if col:
            self.l.player.take_damage(self.contactdmg)



import setuptools
from setuptools import setup

# with open("README.md", "r") as fh:
# 	long_description = fh.read()

setup(    
	name='bluebeam_release', # Name of the PyPI package
	author='Clyde Slichter',
	package_dir={"": "src"},
	packages = setuptools.find_packages(where="src"),
	package_data={
		"bluebeam": ["*.txt","images/*.png", "images/*/*.png", "images/*/*/*.png", "images/*/*/*.PNG" "Sounds/*.wav", "Sounds/*.mod", "Sounds/*.xm", "tilemap_resources/*", "tilemap_resources/*/*", "tilemap_resources/*/*/*", "tilemap_resources/*/*/*/*"]
	},
	author_email='cnslichter@ufl.edu',
	version='2.0.0',
	description='A side scroller, tile map, shooter game', 
	#long_description=long_description,
	#url='https://github.com/josephmbarron/bluebeam',
	install_requires=[
				'pygame>=2.0.1', 
				'PyTMX>=3.25', 
				'wheel'
				'cx-freeze'

	], 
	keywords='pygame, game, shooter', 

	#py_modules=["GlobalVars"],
	python_requires='>=3.8',
	entry_points ={
		'console_scripts': [
			'bluebeam_game = bluebeam.Engine:main',
		],
	},
	include_package_data=True,
)



from bluebeam.objects import Enemy
import pygame


class Target(Enemy):
    def __init__(self, g_, l_, pos_=None):
        super().__init__(g_, l_, pos_)
        self.rect = pygame.rect.Rect(0, 0, 100, 100)
        self.rect.center = self.pos
        self._mhp = 50
        self.hp = self._mhp
        self.idied = 0

    def update(self):
        super().update()
        if self.hp <= 0:
            self.idied = 1

    def render(self):
        pygame.draw.rect(self.l.s,
                         (128 + 127 * (1 - self.idied), 128, 128 + (127 * self.idied), 100),
                         self.rect.copy().move(-self.l.view))
        pygame.draw.rect(self.l.s,
                         (255, 0, 0, 100),
                         pygame.rect.Rect(self.rect.topleft[0], self.rect.topleft[1] - 24, 100, 16).move(-self.l.view))
        pygame.draw.rect(self.l.s,
                         (0, 255, 0, 100),
                         pygame.rect.Rect(self.rect.topleft[0],
                                          self.rect.topleft[1] - 24,
                                          100 * (self.hp / self._mhp),
                                          16).move(-self.l.view))



from bluebeam.objects import Enemy
import pygame, math, random, os
from pygame import Vector2 as vec2

class GammaHand(Enemy):
    def __init__(self, g, l, pos):
        super().__init__(g, l, pos)
        self.rect = pygame.rect.Rect(0, 0, 144, 144)
        self.frect = pygame.rect.Rect(0, 0, 192, 192)
        self.body = None
        self.handtype = vec2(1, 1)
        self.sprites = []
        self.sourceDir = os.path.dirname(os.path.abspath(__file__))
        self.imagedir = os.path.join(os.path.dirname(self.sourceDir), "images/Boss/GammaGenerator")
        self.spritesheet = pygame.image.load(os.path.join(self.imagedir, 'boss03_hands.png')).convert()
        self.load_sprites()

        self.punchstarts = [vec2(), vec2()]
        self.groundlevel = 0

        self.depth = 11

        self.mode = 0
        self.stateclock = 0
        self.bob = 0
        self.boffset = 0
        self._contactdmg = 0

        self.moveto = vec2()
        self.radiatetarg = vec2()
        self.radiatebursts = 10
        self.radiaterange = [vec2(), 0]

        self.corridorclock = 0

        self.emitpoint = self.pos.xy

    def load_sprites(self):
        for i in range(0,2):
            x = i*48
            img = self.load_single((x,0,48,48))
            self.sprites.append(pygame.transform.scale(img,(192,192)))

    def setup(self, body_, right_, bottom_):
        self.body = body_
        self.handtype = vec2(-1 + 2*right_, -1 + 2*bottom_)
        self.rect.center = self.pos = self.body.pos
        self.groundlevel = self.body.groundlevel
        self.punchstarts[0] = vec2(self.body.bounds[0].x, self.body.bounds[0].y)
        self.punchstarts[1] = vec2(self.body.bounds[1].x, self.body.bounds[0].y + 144)
        if bottom_:
            self.boffset = 90

    def update(self):
        self.process()
        self.collision_player()

    def state(self, mode_):
        self.mode = mode_
        self.stateclock = 0
        if mode_ == 2:
            self.radiatebursts = 0
            self.radiate_reposition()
        if mode_ == 3 or mode_ == 5:
            self.moveto = vec2(self.punchstarts[int(0.5 + self.handtype.x/2)].x, self.punchstarts[0].y)
            self.radiatebursts = 0
        if mode_ == 6:
            self.radiatebursts = 0
            self.corridorclock = 0
            self.moveto = vec2(self.body.pos.x + 140 * self.handtype.x, self.punchstarts[0].y)

    def process(self):
        states = {
            0: self.idle,
            1: self.punch,
            2: self.radiate,
            3: self.laser_A,
            4: self.laser_B,
            5: self.corridor_laser,
            6: self.corridor_bullet,
        }
        states[self.mode]()

    def smooth_move(self, target_, speed):
        mag = (target_ - self.pos).magnitude()
        if mag > 0:
            if (target_ - self.pos).magnitude() < math.pow(mag,0.45):
                self.rect.center = self.pos = target_
            else:
                amt = min(speed*self.g.delta_time, math.pow(mag,0.45))
                self.rect.center = self.pos = self.pos + amt * (target_ - self.pos).normalize()

    def idle(self):
        x = 240 + 60*(self.handtype.y==1)
        y = 120 * self.handtype.y - 60 + math.cos(math.radians(self.body.bob + self.boffset))*48
        self.smooth_move(self.body.pos + vec2(x * self.handtype.x, y), 300)
        pass

    def punch(self):
        x = self.punchstarts[int(0.5 + self.handtype.x)].x
        y = self.punchstarts[int(0.5 + self.handtype.y)].y
        if self.stateclock == 0 and self.pos != vec2(x,y):
            self.stateclock = 0
            self.smooth_move(vec2(x,y), 1600)
        else:
            self.stateclock += self.g.delta_time
            if self.stateclock > 2.0:
                self.state(0)
            elif self.stateclock > 0.8:
                self.swing()

    def swing(self):
        offset = ((self.body.phase<=2) and (self.handtype.y == -1)) * 72
        if self.stateclock < 1:
            self._contactdmg = 10
            div = ((self.stateclock - 0.8) / 0.2)
            x = self.punchstarts[int(0.5+self.handtype.x)].x
            y = self.punchstarts[int(0.5+self.handtype.y)].y
            am = vec2(x,y) - self.body.pos
            targ = vec2(192*self.handtype.x, self.body.bounds[1].y-self.body.pos.y-offset)
            self.rect.center = self.pos = self.body.pos + am.slerp(targ, div)
        elif self.stateclock < 1.2:
            div = ((self.stateclock - 1.0) / 0.2)
            am = vec2(192*self.handtype.x, self.body.bounds[1].y-self.body.pos.y-offset)
            targ = vec2(self.body.bounds[1-int(0.5+self.handtype.x)].x - self.body.pos.x, self.body.bounds[1].y-self.body.pos.y -offset - 96)
            self.rect.center = self.pos = self.body.pos + am.lerp(targ, div)
        elif self.stateclock > 1.5:
            x = 240 + 60 * (self.handtype.y == 1)
            y = 120 * self.handtype.y - 60 - offset
            self.smooth_move(self.body.pos + vec2(x * self.handtype.x, y), 650)
            self._contactdmg = 0
        elif self.stateclock > 2:
            self.state(0)

    def radiate(self):
        if self.pos != self.moveto:
            self.stateclock = 0
            self.smooth_move(self.moveto, 1600)
            self.radiate_genangles(45)
        else:
            self.stateclock += self.g.delta_time
            if self.stateclock >= 0.3 and self.radiatebursts == 0:
                self.radiate_genangles(45)
                self.radiatebursts = 1
            elif self.radiatebursts >= 1 and self.stateclock >= 0.45 - 0.03*(3-self.body.phase):
                self.stateclock = 0
                self.radiate_fire()
                self.radiatebursts += 1
                if (self.radiatebursts % (3+(3-self.body.phase))) == 1:
                    if self.radiatebursts == 10 + 3*(3-self.body.phase):
                        self.state(0)
                    else:
                        self.radiate_reposition()
                        self.radiate_genangles(45)

    def radiate_reposition(self):
        self.rect.center = self.pos
        #self.moveto = self.pos
        self.moveto = vec2(random.randint(self.body.bounds[0].x + 144, self.body.bounds[1].x - 144), random.randint(self.body.bounds[0].y + 144, self.body.bounds[1].y - 288))

    def radiate_genangles(self, spread):
        plrang = (self.l.player.pos - self.pos).normalize()
        self.radiaterange[0] = plrang.rotate(-spread)
        self.radiaterange[1] = spread

    def radiate_fire(self):
        pass
        for i in range(0,3 + int(0.5*(3-self.body.phase))):
             rand = random.random() * (self.radiaterange[1]*2)
             ang = self.radiaterange[0].rotate(round(rand/5)*5)
             bullet = self.l.create("EnemyBullet", vec2(self.pos.x, self.pos.y))
             bullet.speed = ang * 650
             bullet.depth = 24

    def laser_A(self):
        if self.pos != self.moveto and self.radiatebursts % 2 == 0:
            self.smooth_move(self.moveto, 1200)
            self.emitpoint = self.pos
            if self.pos == self.moveto:
                self.radiatebursts += 1
                self.laser_A_fire()
            pass
        else:
            a = (self.body.bounds[1].x - self.body.bounds[0].x) / (450 + 30*(3-self.body.phase))
            self.stateclock += self.g.delta_time
            if self.stateclock > 0.7 + a + 0.4:
                self.state(0)
            elif self.stateclock > 0.7 + a:
                self.pos = vec2(self.body.bounds[int(0.5 - self.handtype.x/2)].x + 192 * self.handtype.x, self.moveto.y)
            elif self.stateclock > 0.7:
                self.pos = self.moveto.lerp(vec2(self.body.bounds[int(0.5 - self.handtype.x/2)].x + 192 * self.handtype.x, self.moveto.y),min(1.0,(self.stateclock-0.7)/a))
            self.emitpoint = self.pos
            self.rect.center = self.pos
        pass

    def laser_A_fire(self):
        las = self.l.create("Laser", self.pos.xy)
        las.warmup = 0.7
        las.width = 96
        las.depth = 16
        las.dir = 1
        las.lifespan = (self.body.bounds[1].x - self.body.bounds[0].x) / (400 + 50*(3-self.body.phase)) + 0.2
        las._emitter = self

    def laser_B(self):
        pass

    def corridor_laser(self):
        if self.radiatebursts == 0 and self.pos != self.moveto:
            self.smooth_move(self.moveto, 750)
            if self.pos == self.moveto:
                self.radiatebursts = 1
                self.corridor_laser_fire()
                pass
        else:
            self.stateclock += self.g.delta_time
            if self.stateclock > 16:
                self.state(0)
            if self.stateclock > 4:
                self.pos = self.rect.center = vec2(self.body.pos.x + 256 * self.handtype.x, self.moveto.y)
            elif self.stateclock > 1:
                self.pos = self.rect.center = self.moveto.lerp(vec2(self.body.pos.x + 256 * self.handtype.x, self.moveto.y), (self.stateclock - 1) / 3)
            self.emitpoint = self.pos
        pass

    def corridor_laser_fire(self):
        las = self.l.create("Laser", self.pos.xy)
        las.warmup = 1
        las.width = 150
        las.dir = 1
        las.depth = 16
        las.lifespan = 13
        las._emitter = self

    def corridor_bullet(self):
        if self.radiatebursts == 0 and self.pos != self.moveto:
            self.smooth_move(self.moveto, 300)
            if self.pos == self.moveto:
                self.radiatebursts = 1
                pass
        else:
            self.stateclock += self.g.delta_time
            self.corridorclock += self.g.delta_time
            if self.stateclock > 13:
                self.state(0)
            elif self.corridorclock > 2.5 and self.radiatebursts == 1:
                self.corridor_bullet_fire()
                self.corridorclock = 0
                self.radiatebursts = 2
            elif self.radiatebursts >= 2 and self.corridorclock > 1.0 - (self.body.phase == 1)*0.05 - min(0.15, (self.radiatebursts - 2)/50):
                self.corridor_bullet_fire()
                self.corridorclock = 0
                self.radiatebursts += 1

    def corridor_bullet_fire(self):
        for i in range(0,3):
            bul = self.l.create("EnemyBullet",self.pos.xy)
            bul.speed = vec2(math.cos(math.radians(90+5*i*self.handtype.x)),math.sin(math.radians(90+5*i*self.handtype.x))) * 750
            bul.depth = 16
        pass

    def death(self):
        boom = self.l.create("Explosion",self.pos.xy)
        boom.radius(480)
        boom.dmg = 0
        boom.depth = 64
        self.l.uncreate(self.instid)

    def render(self):
        type=0
        self.frect.center = self.rect.center
        if self.body.explo > 0 or self.mode == 2 or self.mode == 3 or self.mode == 5 or self.mode == 6:
            type=1
        img = self.sprites[type]
        if self.handtype.x==1:
            img = pygame.transform.flip(img,True,False)
        self.l.s.blit(img,self.frect.copy().move(-self.l.view))


import pygame
import pytmx
import os
from pytmx.util_pygame import load_pygame

from bluebeam.objects import Block
from math import floor, ceil

class TileMap:
    def __init__(self, globals, level, filename="TestMap.tmx"):
        self.globals = globals
        self.level = level
        #self.sourceDir = os.getcwd()
        self.sourceFileDir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        self.map_dir = os.path.join(self.sourceFileDir, "bluebeam/tilemap_resources/")
        #print("hey", self.sourceFileDir)
        #self.tiled_map = load_pygame("tilemap_resources/" + filename)
        self.tiled_map = load_pygame(os.path.join(self.map_dir, filename))

        self.width  = self.tiled_map.width
        self.height = self.tiled_map.height

        self.images = {-500:None}

        # Used as a way to calculate the corners of the screen given camera position
        (screen_width, screen_height) = globals.screen_size
        self.screen_size = pygame.math.Vector2(screen_width, screen_height)

        self._load_gid_map()
        self._load_block_map()

    # Create 2d array of Image Placement
    def _load_gid_map(self):
        # Initialize empty 2d list of tiles
        self.tiles = [[0 for j in range(self.tiled_map.height)] for i in range(self.tiled_map.width )]

        # Populate 2d list of tiles
        for layer in self.tiled_map.layers:
            if isinstance(layer, pytmx.TiledTileLayer):
                for x, y, gid in layer:
                    if (gid):
                        self.tiles[x][y] = gid
                        if (gid not in self.images):
                            self._load_image(gid)

    # Load tile images used in the tilemap
    def _load_image(self, gid):
        print("loading tile w/ gid #",gid)
        img = self.tiled_map.get_tile_image_by_gid(gid)
        if (img != None):
            img = img.convert()
            self.images[gid] = img

    # Create list of Collision Blocks
    def _load_block_map(self):
        for tile_object in self.tiled_map.objects:
            if tile_object.name == "wall":
                pos = (tile_object.x, tile_object.y)
                block = Block(self.globals, self.level, pos, tile_object.width, tile_object.height)
                self.level.append(block)

    def load_player_pos(self):
        for tile_object in self.tiled_map.objects:
            if(tile_object.name == "player"):
                return pygame.math.Vector2(tile_object.x, tile_object.y)

    def load_pickup_info(self):
        pickups = []
        for tile_object in self.tiled_map.objects:
            if(tile_object.name == "pickup"):
                pos = pygame.math.Vector2(tile_object.x+50, tile_object.y+50)
                tup = (tile_object.type, pos)
                pickups.append(tup)
        return pickups

    def load_enemy_info(self):
        enemies = []
    
        for tile_object in self.tiled_map.objects:
            if (tile_object.name == "enemy"):
                pos = pygame.math.Vector2(tile_object.x+50, tile_object.y+50)
                enemies.append( (tile_object.type, pos) )
        return enemies

    # Draw the tiles on the screen
    def render(self, screen, camera_offset):
        (xmin, xmax, ymin, ymax) = self.calc_render_indices(camera_offset)
        for i in range(xmin, xmax):
            for j in range(ymin, ymax):
                gid = self.tiles[i][j]
                try:
                    tile = self.images[gid]
                    true_pos = pygame.math.Vector2(i * 64, j * 64)
                    screen_pos = true_pos - camera_offset
                    screen.blit(tile, screen_pos)
                except:
                    print("ERROR: BROKE ON GID: " + str(gid))

    def calc_render_indices(self, camera_offset):
        top_left_coords  = camera_offset
        bot_right_coords = camera_offset + self.screen_size

        # Convert screen Coordinates to Valid Tilemap coordinates
        xmin = max( floor(top_left_coords.x / 64), 0           )
        xmax = min( ceil(bot_right_coords.x / 64), self.width  )
        ymin = max( floor(top_left_coords.y / 64), 0           )
        ymax = min( ceil(bot_right_coords.y / 64), self.height )

        return (xmin, xmax, ymin, ymax)
